/* msggen - a small diagnostic utility that does very quick
 * syslog() calls.
 *
 * Copyright 2008-2014 Rainer Gerhards and Adiscon GmbH.
 *
 * This file is part of rsyslog.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *       -or-
 *       see COPYING.ASL20 in the source distribution
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <syslog.h>

int main(int __attribute__((unused)) argc, char __attribute__((unused)) * argv[]) {
    int i;

    openlog("msggen", 0, LOG_LOCAL0);

    for (i = 0; i < 10; ++i) syslog(LOG_NOTICE, "This is message number %d", i);

    closelog();
    return 0;
}
